/*
ID-TIMER.C Copyright (c) 2004-2008, Dale Botkin / N0XAS

All rights reserved.  You may use and alter this program for
your own personal use or for club projects.  You may not sell
programs derived from, or devices using this program, without
written consent from the author.  Contact me via email to
n0xas@arrl.net, or by Snail Mail for permission.  I'm a
pretty reasonable guy.

V1.0:   2/10/2004   Initial release, seems to work OK.
V1.1:   2/12/2004   Added CW and PTT outputs, cleaned up code
V1.2:   2/13/2004   Added repeater controls
V1.3:   2/15/2004   Code cleanup, added repeater mode, longer messages
V1.4:   7/22/2004   Add test message, HS instead of INTRC_IO, '=' sends BT
V2.0:   11/16/2004  Change to HS, 4MHz xtal.  Moved PTT & CW to A0 & A1.
        12/02/2004  Fixed numerous bugs.  Got CW keyboard mode working.
                    Changed message & buffer size to 65.
        12/24/2004  PTT now goes high when timeout occurs in timer mode.
        01/18/2005  Added courtesy beep for repeater mode, one step closer
                    to a repeater controller.
V2.1:   02/09/2005  Fixed & tweaked courtsey beep, tested OK.  Released
                    to production.
V2.2:   03/08/2005  Changed repeater mode to ID 3 seconds after the first
                    transmission ends, instead of 30.
V2.3:   10/22/2005  Changed behavior of first repeater ID to only fire
                    if it's been more than id_time since the last ID.
                    This fixed a problem reported by VK8RR where we'd ID
                    twice in a row in repeater mode.
                    Added repeater-mode beacon time.
                    Allow 0 for yellow & blink times.
                    SPACE will delete ID message.
V2.4:   11/21/2005  Added second message for BEACON mode ID in repeater mode
V2.5:   02/22/2006  Fixed bug with beacon time & no CW ID.  This was reported
                    by G7RPG, changing settings may cause the device to start
                    beeping & hang with no KBD response.
                    Added PTT hang time to improve rpt mode, thanks G7RPG!
                    Changed to PWM audio out, delay_ms for Morse timing
                    Added variable freq. courtesy beep and Morse ID
                    Audio freq. entry is now by Hz instead of delay values
        03/01/2006  Added PTT watchdog timer
V2.6:   07/22/2006  Tweak to prevent timing stops on serial port noise.
                    Added multiplier for courtesy beep length.
                    Added :nnn: to send carrier for nnn seconds
                    Released to production 8/17/2006
V2.7:   11/03/2006  Altered test to only send '/' & MSG2 if MSG2 exists
                    Fix for random beacon in ID mode - ISR was altering
                    long int variables while they were being used.
V2.8:   09/05/2007  Changed behavior of START and INHIBIT.  START now
                    acts as a "soft" inhibit which will hold off ID until it
                    goes inactive OR the timer expires.  INH is a "hard"
                    inhibit that will hold off as long as it's active.  Also
                    cleaned up code, changed START to START throughout.
        09/17/2007  Changed short_atol() to make it dedicated to processing
                    tbuffer; this saves ~20 words.
V2.9:   02/24/2008  * Consolidated two functions into one to get all messages
                      with the same function.  Gained ~140 words.
                    * New functions for getting tones and timers, major code
                      space shrink.
                    * Added ability to send an alternate ID message.
                    * TEST now sends the version number, then all messages in
                      sequence.
                    * TEST input now switches to ALT_MSG use after initial
                      power on.
                    * In timer mode, timer beeps 10x then resets automatically.
        03/04/2008  * Added variable delay for courtesy beep
                    * PTT goes active as soon as START is seen, regardless of
                      PTT hang time setting
V2.10:  04/03/2008  * TOT watchdog resets as soon as START goes inactive
                    * send_msg() does not drop PTT after :nnn:
                    * Better control over LED state, especially blinking red
        05/12/2008  * Use RA2 for beacon indicator.  Should be high when beacon
                      is sent (regardless of message number used), low the rest
                      of the time.

*/

#include <16F648a.h>
#device *=16
#fuses NOWDT,HS,NOLVP,NOCPD,NOPROTECT,MCLR,PUT,NOBROWNOUT
#use delay(clock=4000000)
#use rs232(baud=9600,parity=N,xmit=PIN_B2,rcv=PIN_B1,bits=8,errors)
#use standard_io(B)
#use fast_io(A)
#include <string.h>
#define TEST    PIN_B0
#define RXD     PIN_B1
#define TXD     PIN_B2
#define AUDIO   PIN_B3
#define LED_RED PIN_B4
#define LED_GRN PIN_B5
#define INHIBIT PIN_B6
#define START   PIN_B7
#define PTT     PIN_A0
#define CW      PIN_A1
#define BCN_IND PIN_A2
#define ALT_MSG PIN_B0
#byte   INTCON = 0x0B
#byte   PIR1   = 0x0C
#define IDSAVE      0
#define BCNSAVE     2
#define YELSAVE     4
#define BLSAVE      6
#define AUDSAVE     8
#define BEEPSAVE    10
#define WDSAVE      12
#define AUTOSAVE    14
#define WPMSAVE     15
#define RPTSAVE     16
#define HANGSAVE    17
#define MULTSAVE    18
#define CDELAYSAVE  19
#define MSG_START   20
#define MSG_LEN     64
#define ETX         0x03
#define MAJ_REV 2
#define MIN_REV 10
#define MSG2_START  MSG_START + MSG_LEN
#define MSG3_START  MSG2_START + MSG_LEN
#define ID_DELAY    2000
#define ONESEC      1000

#rom 0x2100 = { 0x58,   // ID time (lo byte)
                0x02,   // ID time (hi byte)
                0x00,   // Beacon time (lo byte)
                0x00,   // Beacon time (hi byte)
                60,     // Yellow time (lo byte)
                0,      // Yellow time (hi byte)
                30,     // Blink time (lo byte)
                0,      // Blink time (hi byte)
                82,     // CW tone (lo byte)
                0,      // CW tone (hi byte)
                82,     // Courtesy beep tone (lo byte)
                0,      // Courtesy beep tone (hi byte)
                0,      // Watchdog time (lo byte)
                0,      // Watchdog time (hi byte)
                'N',    // Auto CW ID
                15,     // CW speed
                'N',    // Repeater mode
                0,      // PTT hang time
                1,      // CBEEP multiplier
                5,      // CBEEP delay time
                0}      // Start of ID message

#rom 0x2100 + MSG2_START = {0}
#rom 0x2100 + MSG3_START = {0}

char buffer[65], tbuffer[65];
char auto_id, speed, is_rpt, since_first, ptt_count, ptt_hangtime, cbeep_mult;
char cbeep_delay;
unsigned long int seconds, id_time, yel_time, bl_time, dit_length, since_id;
unsigned long int bcn_time, cbeep_tone, aud_val, toneval, watchdog, ptt_timer;
int32 count;

struct bits {
    short red_on;
    short green_on;
    short green_status;
    short green_blink;
    short red_blink;
    short ptt_active;
    short ptt_timeout;
    short start_timer;
    short sent_cbeep;
    short no_seconds;
} flags;

const char morsechar[] = {
    0x73,   // ,
    0x31,   // - (BT)
    0x6A,   // .
    0x29,   // /
    0x3F,   // 0
    0x3e,   // 1
    0x3c,   // 2
    0x38,   // 3
    0x30,   // 4
    0x20,   // 5
    0x21,   // 6
    0x23,   // 7
    0x27,   // 8
    0x2F,   // 9
    0x01,   // : (nothing sent)
    0x2A,   // ; (AR)
    0x01,   // < (nothing sent)
    0x31,   // = (BT)
    0x68,   // > (SK)
    0x4C,   // ?
    0x56,   // @
    0x06,   // A
    0x11,   // B
    0x15,   // C
    0x09,   // D
    0x02,   // E
    0x14,   // F
    0x0B,   // G
    0x10,   // H
    0x04,   // I
    0x1E,   // J
    0x0D,   // K
    0x12,   // L
    0x07,   // M
    0x05,   // N
    0x0F,   // O
    0x16,   // P
    0x1B,   // Q
    0x0A,   // R
    0x08,   // S
    0x03,   // T
    0x0C,   // U
    0x18,   // V
    0x0E,   // W
    0x19,   // X
    0x1D,   // Y
    0x13    // Z
};

void ptt_on(void);
void ptt_off(void);
void char get_ee(char address);
#separate
void send_it(char x);
#separate
void cw_id(char msg);
void word(void);
#separate
void transmit(char length);
#separate
void send_morse(char data);
#separate
void process_timers(void);


/*
Timer0 is used to keep track of the time since reset.  We have 1 million
cycles per second, and T0 interrupts every 65536 uS.  We add 65536 to
the counter every interrupt.  If it rolls over 1 million we add one to
the seconds count, subtract a million and keep going.  This way we can
have up to 65ms of jitter, but the long term accuracy will be as good as
our clock.  Over a nominal 10 minute period, it's good enough without
trimming the oscillator.
*/
#int_TIMER0
void TIMER0_isr() {
    count += 4096;
    if(count > 1000000) {
        count -= 1000000;
        if(is_rpt != 'Y' || flags.start_timer) {
            if(seconds > 0) {
                flags.no_seconds = FALSE;
                seconds--;
            } else {
                flags.no_seconds = TRUE;
            }
        }

        if(since_id < 0xFFFF) {
            since_id++;
        }

        if(ptt_count > 0) {
            ptt_count--;
        }

        if(ptt_timer > 0) {
            if(!--ptt_timer) {
                flags.ptt_timeout = TRUE;
            }
        }

        if(yel_time && (seconds < yel_time)) {
            flags.red_on = TRUE;                // LED should be yellow
            flags.green_on = TRUE;
            flags.green_blink = FALSE;
        }

        if(bl_time && (seconds < bl_time)) {
            flags.red_on = TRUE;                // LED should be yellow
            flags.green_on = TRUE;
            flags.green_blink = TRUE;           // and blinking
        }

        output_low(LED_RED);
        if(flags.red_on) {
            output_high(LED_RED);
        }

        output_low(LED_GRN);
        if(flags.green_on) {
            output_high(LED_GRN);
        }

        if(flags.green_blink) {
            if(flags.green_status) {
                output_high(LED_GRN);
                flags.green_status = FALSE;
            } else {
                output_low(LED_GRN);
                flags.green_status = TRUE;
            }
        }
    }
}

/*
Get an integer value from EEPROM.  This function reads ASCII values
from EEPROM in the form of :nnn: and returns the integer value
represented by nnn.
*/
int int_from_ee(char n) {
    int result = 0, x;

    do {
        x = get_ee(n++);
        if(x != ':') {
            result *= 10;
            result += (x - 0x30);
        }
    } while(x != ':');
    return result;
}

/*
Reads a string from the serial port, with echo.
*/
void get_string(char* s, int max) {
   int len;
   char c;

   --max;
   len=0;
   do {
     c=toupper(getc());
     if(c==8) {  // Backspace
        if(len>0) {
          len--;
          putc(c);
          putc(' ');
          putc(c);
        }
     } else if ((c>=' ')&&(c<='~'))
       if(len<max) {
         s[len++]=c;
         putc(c);
       }
   } while(c!=13);
   s[len]=0;
}

/*
A little irritating, but for some reason multiple uses of the
write_eeprom() function are not very efficient at all.  Much
better at this point to put the function call in one place with
a wrapper.  Ugly, but it works better this way.  This may be
fixed in a later compiler version.
*/
void save(char address, char data) {
    write_eeprom(address,data);
}

/*
Same deal, but for EEPROM reads.
*/
char get_ee(char address) {
    return read_eeprom(address);
}

/*
Turns PTT on.  Made this a function to save space.
*/
void ptt_on(void) {
    output_high(ptt);
    if(!flags.ptt_active) {
        flags.ptt_active = TRUE;
        ptt_timer = watchdog;
    }
}

/*
Turns PTT off.  Made this a function to save space.
*/
void ptt_off(void) {
    output_low(PTT);
    flags.ptt_active = FALSE;
    ptt_timer = 0;
}


/*
Generate sidetone audio & keying signal.  The AUDIO pin gets a square
wave tone, while the CW line can be used to key a CW transmitter if
needed.  This would be useful for a beacon or whetever.
*/
#separate
void transmit(char length) {
    char x, delay;
    long y;

    if(length == 0)
        return;

    dit_length = 1200 / speed;
    setup_timer_2(T2_DIV_BY_16,toneval,1);
    set_pwm1_duty(toneval / 4);
    output_high(AUDIO);
    output_high(CW);
    for(x=0;x<length;x++) {
        delay_ms(dit_length);
    }
    output_low(CW);
    output_float(AUDIO);
    // Now delay one dit length for element spacing
    delay_ms(dit_length);
}

/*
Sends a Morse character.  Data is shifted out of the LSB,
0 for dit and 1 for dah.  Character has a trailing 1, so
when the data is shifted down to 0x01 we know we're done.
*/
#separate
void send_morse(char data) {
    char x;

    while(data != 0x01) {
        // The following asm code is the shortest path to transmitting
        // a dit for a zero in the LSB of data, and a dah for a 1 bit in
        // the LSB of data.  Assembly here was much more efficient than
        // compiled code.
        #asm
        clrf &x         // Clear x
        bcf 03.0        // Clear carry bit
        rrf &data, f    // Shift bit out of data
        rlf &x, f       // Shift bit into x
        // x now has 0 for dit, 1 for dah
        bsf 03.0        // Set carry bit
        rlf &x, f       // Shift bit into x
        // x now has 01 for dit, 11 for dah
        #endasm
        transmit(x);
    }
    for(x=0;x<2;x++) {
        delay_ms(dit_length);
    }
}

/*
This is an adaptation of the C atol() function, pared down to
accept only positive numbers, no sign, 0 to 9 only.  Saves a lot
of space that way.
*/
long short_atol(void) {
    long result = 0;
    char c, index = 0;
    char *s = tbuffer;

    if(s)
        c = s[index++];

    while (c >= '0' && c <= '9') {
        result = 10*result + (c - '0');
        c = s[index++];
    }
    return(result);
}

/*
word() simply delays six dit times, a proper Morse code word space.
*/
void word(void) {
    char x;

    for(x=0;x<5;x++) {
        delay_ms(dit_length);
    }
}

/*
Sends a Morse character.  If the character is a space, sends a
space.  This function checks to make sure the character passed
is in the Morse table; if it's not, the character is ignored.
*/
#separate
void send_it(char x) {
    char y;

    if(x == ' ') {
        word();
        return;
    }
    y = 1;
    if(x >= ',' && x <= 'Z')
        y = morsechar[x - ','];
    send_morse(y);
}

/*
Sends an ID message, manages PTT, LED color, etc. as needed.

INPUT =     message number to send, or 0 to send whichever
            message is appropriate depending on timer status and
            which message(s) are set.
OUTPUT =    None.
*/
#separate
void cw_id(char message) {
    char n, x;
    if(message) {
        n = message;
    } else {
        if(flags.no_seconds) {
            if (!input(ALT_MSG) && get_ee(MSG3_START)) {
                n = MSG3_START;
            } else {
                n = MSG_START;
            }
        // Here we have a beacon event
        } else {
            output_high(BCN_IND);
            if(get_ee(MSG2_START)) {
                n = MSG2_START;
            } else {
                if (!input(ALT_MSG) && get_ee(MSG3_START)) {
                    n = MSG3_START;
                } else {
                    n = MSG_START;
                }
            }
        }
    }
    flags.red_on = TRUE;                // LED should be red
    flags.green_blink = FALSE;
    flags.green_on = FALSE;
    while(!input(INHIBIT))
        ;
    ptt_on();
    delay_ms(ID_DELAY);
    while(x = get_ee(n++)) {
        if(x != ':') {
            send_it(x);
        } else {
            x = int_from_ee(n);
            setup_timer_2(T2_DIV_BY_16,toneval,1);
            set_pwm1_duty(toneval / 4);
            output_high(AUDIO);
            output_high(CW);
            do {
                delay_ms(ONESEC);
            } while(--x);
            output_low(CW);
            output_float(AUDIO);
            do {
                (x = get_ee(n++));
            } while(x != ':');
        }
    }
    delay_ms(100);
    if(input(START)) {
        ptt_off();
    }
    flags.red_on = FALSE;               // LED should be green
    flags.green_on = TRUE;
    flags.green_blink = FALSE;
    flags.start_timer = FALSE;
    since_id = 0;
    seconds = id_time;
    flags.no_seconds = FALSE;
    output_low(BCN_IND);
}

/*
Check for PTT timeout (TOT).  If the watchdog flag is set,
send TOT and ID, then wait for START to drop.
*/
void check_for_timeout(void) {
    if(flags.ptt_timeout) {
        send_it('T');
        send_it('O');
        send_it('T');
        word();
        seconds = 0;
        flags.no_seconds = TRUE;
        cw_id(0);
        ptt_off();
        while(!input(START))
            ;
        flags.ptt_timeout = FALSE;
    }
}


/*
Limits a long int to the range of 250-3500.
*/
long check_range(long data) {
    if(data > 5000)
        data = 5000;
    if(data < 250)
        data = 250;
    return data;
}

void get_response(void) {
    while(!kbhit())
        ;
    get_string(tbuffer, 6);
}

void process_long(long *value) {
    long scratch;

    if(tbuffer[0]) {
        scratch = short_atol();
        *value = scratch;
    }
}

void save_long (char address, long value) {
        save(address, *(&value));
        save(address+1, *(&value+1));
}

long fetch_long (char address) {
    long x;

    x = get_ee(address);
    x += (get_ee(address+1) * 256);
    return x;
}

#separate
void get_stuff_timer(char address, long timer, char bytes) {
    printf(" time (%lu): ",timer);
    get_response();
    process_long(&timer);
    if(bytes > 0) {
        save_long(address,timer);
    } else {
        save(address, timer);
    }
}

#separate
void get_stuff_tone(char address, long tone) {
    long scratch;

    printf(" tone (%lu): ", 62500 / (tone + 1));
    get_response();
    if(tbuffer[0]) {
        scratch = short_atol();
        scratch = check_range(scratch);
        tone = (62500 / scratch) - 1;
        save_long(address, tone);
    }
}

#separate
void get_stuff_msg(char msgno) {
    char x;
    for(x=0;x<MSG_LEN;x++)
        buffer[x] = read_eeprom(msgno + x);
    printf(" Msg (%S): ",buffer);
    while(!kbhit())
        ;
    get_string(tbuffer, 64);
    if(tbuffer[0]) {
        if(tbuffer[0] == 0x20) {
            tbuffer[0] = 0;
            bcn_time = 0;
        }
        for(x = 0; x < MSG_LEN; x++) {
            save(msgno + x, tbuffer[x]);
            buffer[x] = tbuffer[x];
        }
    }
}

#separate
void get_stuff_auto_id(void) {
    printf("\r\nAuto CW ID (%C): ",auto_id);
    get_response();
    if(tbuffer[0]) {
        if(tbuffer[0] == 'Y') {
            auto_id = 'Y';
        } else {
            auto_id = 'N';
            bcn_time = 0;
        }
    }
    save(AUTOSAVE,auto_id);
}

#separate
void get_stuff_cw(void) {
    long scratch = 0;

    printf("\r\nCW Speed (%U): ",speed);
    get_response();
    scratch = short_atol();
    if(scratch) {
        speed = scratch;
        save(WPMSAVE, speed);
    }

    printf("\r\nID audio");
    get_stuff_tone(AUDSAVE, aud_val);

}

#separate
void get_stuff_rptmode(void) {
    long scratch = 0, fq = 0;
    char cscratch;

    printf("\r\nRepeater mode (%C): ",is_rpt);
    get_response();
    if(tbuffer[0]) {
        if(tbuffer[0] == 'Y') {
            is_rpt = 'Y';
        } else {
            is_rpt = 'N';
            bcn_time = 0;
            save_long(BCNSAVE,bcn_time);
        }
        save(RPTSAVE,is_rpt);
    }
    if(is_rpt == 'Y') {
        if(cbeep_tone) {
            fq = 62500 / (cbeep_tone + 1);
        }

        printf("\r\nCourtesy beep");
        get_stuff_tone(BEEPSAVE, cbeep_tone);

        printf("\r\nCourtesy beep delay");
        get_stuff_timer(CDELAYSAVE, cbeep_delay, 0);

        printf("\r\nCourtesy beep mult (%u): ", cbeep_mult);
        get_response();
        if(tbuffer[0]) {
            scratch = short_atol();
            if((scratch >= 0) && (scratch < 100)) {
                cbeep_mult = scratch;
            }
            save(MULTSAVE, cbeep_mult);
        }


        printf("\r\nBeacon");
        get_stuff_timer(BCNSAVE, bcn_time, 1);

        printf("\r\nPTT hang");
        get_stuff_timer(HANGSAVE, ptt_hangtime, 0);

        printf("\r\nPTT max");
        get_stuff_timer(WDSAVE, watchdog, 1);

    }
}

/*
Here we get user parameters from the serial port and save them in
EEPROM.  Not the least bit elegant, but it's a very simple device.
Hams can figure out how to follow directions (we hope).
*/
void get_stuff(void) {
    char x = 0;
    long scratch;

    output_low(LED_GRN);
    output_high(LED_RED);
    printf("\r\nID");
    get_stuff_timer(IDSAVE, id_time, 1);
    printf("\r\nYellow");
    get_stuff_timer(YELSAVE, yel_time, 1);
    printf("\r\nBlink");
    get_stuff_timer(BLSAVE, bl_time, 1);
    printf("\r\nSPACE will delete ID string.\r\nID");
    get_stuff_msg(MSG_START);
    printf("\r\nBeacon");
    get_stuff_msg(MSG2_START);
    printf("\r\nAlternate");
    get_stuff_msg(MSG3_START);
    get_stuff_auto_id();
    get_stuff_cw();
    get_stuff_rptmode();
    reset_cpu();
}

void main() {
    char t, x, y, n, nobeep;
    unsigned long int tmp_sec1, tmp_sec2, tmp_since1, tmp_since2;

    setup_comparator(NC_NC_NC_NC);
    port_b_pullups(TRUE);
    output_low(LED_RED);
    output_low(LED_GRN);
    output_low(PTT);
    output_low(CW);
    setup_timer_0(RTCC_INTERNAL|RTCC_DIV_16);
    setup_timer_1(T1_INTERNAL|T1_DIV_BY_8);
    setup_comparator(NC_NC_NC_NC);
    setup_vref(FALSE);
    flags = 0;
    enable_interrupts(INT_TIMER0);
    id_time = fetch_long(IDSAVE);
    yel_time = fetch_long(YELSAVE);
    bl_time = fetch_long(BLSAVE);
    bcn_time = fetch_long(BCNSAVE);
    watchdog = fetch_long(WDSAVE);
    cbeep_tone = fetch_long(BEEPSAVE);
    aud_val = fetch_long(AUDSAVE);
    auto_id = get_ee(AUTOSAVE);
    speed = get_ee(WPMSAVE);
    is_rpt = get_ee(RPTSAVE);
    ptt_hangtime = get_ee(HANGSAVE);
    cbeep_mult = get_ee(MULTSAVE);
    cbeep_delay = get_ee(CDELAYSAVE);
    setup_ccp1(CCP_PWM);
    output_float(AUDIO);
    toneval = aud_val;
    setup_timer_2(T2_DIV_BY_16,toneval,1);
    seconds = id_time;                  // Reset timer
    count = 0;
    flags = 0;
    since_first = 0;
    since_id = id_time;
    output_a(0);
    set_tris_a(0);
    enable_interrupts(global);
    printf("\n\r\nN0XAS ID-O-Matic V%u.%u\n\r",MAJ_REV,MIN_REV);
    flags.green_on = TRUE;

    /*
    See if the user is pressing the test button when power is applied.
    Note that after V2.9, the TEST input is now also the ALT_MSG
    input.  After initial power on, the input is used to select an
    alternate ID message.
    */
    if(!input(TEST)) {
        printf(send_it,"V%u.%u ",MAJ_REV, MIN_REV);
        send_it('/');
        word();
        if(get_ee(MSG_START) || get_ee(MSG2_START) || get_ee(MSG3_START)) {
            cw_id(MSG_START);
            if(get_ee(MSG2_START)) {
                word();
                send_it('/');
                cw_id(MSG2_START);
            }
            if(get_ee(MSG3_START)) {
                word();
                send_it('/');
                cw_id(MSG3_START);
            }
        }
    }

    while(1) {
        /*
        See if we need to talk to the user...
        */
        if(kbhit()) {                           // Do we have serial input?
            t = 0;
            x = getc();
            if((x == '\r') || (x == '\n')) {    // See if it's CR or LF
                while(t++ < 20) {
                    delay_ms(250);
                    if(kbhit()) {
                        x = getc();
                        if((x == '\r') || (x == '\n')) {    // See if it's CR or LF
                            get_stuff();
                        }
                    }
                }
            }
        }

        /*
        Handle repeater mode.  When the START input is seen, we set
        the ID trigger.  If this is the first time we've seen START
        active in more than the timeout interval, we set the timeout to
        10 seconds.  If courtesy beep is on, we wait half a second after
        START ends and send the courtesy beep.  If we have a PTT hang
        time set, turn PTT on as soon as START is seen, and leave it on
        until ptt_hangtime after it drops.
        */
        if(is_rpt == 'Y') {
            if(!input(START)) {
                ptt_on();
                flags.green_blink = TRUE;
                flags.start_timer = TRUE;
                flags.sent_cbeep = FALSE;
                /*
                Here we set the seconds counter to 10 if it's been longer
                than the normal ID interval since we sent the ID - in other
                words, if it's the first repeater use after it's been idle for
                more than 10 minutes (or whatever the interval is set for).
                We also set since_id to zero so we're sure to ID at the end
                of the normal interval, whether the START input is active or
                not.
                */
                if(since_id > id_time) {
                    seconds = 10;
                    since_id = 0;
                }
                disable_interrupts(GLOBAL);
                ptt_count = ptt_hangtime;
                enable_interrupts(GLOBAL);
                check_for_timeout();
            } else {
                if(flags.ptt_active) {          // PTT is on, but START is off
                    if(cbeep_mult && !flags.sent_cbeep) {
                        for(x = 0; x < cbeep_delay; x++) {
                            delay_ms(100);      // Wait 100ms
                            if(!input(START)) { // See if START is active again
                                nobeep = TRUE;  // Yes = no courtesy beep or
                                                // watchdog reset
                                break;
                            } else {
                                nobeep = FALSE; // Otherwise we reset the TOT
                                                // watchdog
                            }
                        }
                        if(nobeep == FALSE) {   // Do the courtesy beep
                            toneval = cbeep_tone;
                            transmit(cbeep_mult);
                            toneval = aud_val;
                            disable_interrupts(GLOBAL);
                            ptt_timer = watchdog;       // Reset watchdog
                            ptt_count = ptt_hangtime;   // Set hang time
                            enable_interrupts(GLOBAL);
                            flags.sent_cbeep = TRUE;
                        }
                    }
                }
                if(!ptt_count) {
                    ptt_off();
                    flags.sent_cbeep = FALSE;
                }
            }
        }

        if(auto_id == 'N') {                    // Simple timer mode
            if(flags.no_seconds) {
                output_low(LED_GRN);
                output_high(LED_RED);
                for(y = 0; y < 10; y++) {
                    output_high(PTT);
                    transmit(speed / 2);
                    //                    disable_interrupts(GLOBAL);
                }
                output_low(PTT);
                reset_cpu();
            }
        } else {
            if(flags.no_seconds || ((bcn_time > 0) && (since_id >= bcn_time))) {
                flags.green_blink = FALSE;
                output_low(LED_GRN);
                output_high(LED_RED);
                /* Hold off until START drops OR we hit the normal ID time. */
                while(!input(START) && (since_id < id_time)) {
                    check_for_timeout();
                }
                cw_id(0);
                delay_ms(ID_DELAY);
                output_low(LED_RED);
                output_high(LED_GRN);
                flags.start_timer = FALSE;
            }
        }
    }
}
